#include <utils/Serializer.hpp>

#ifdef PUGS_HAS_COSTO

// MESH node position
void
Serializer::mesh(const std::shared_ptr<const MeshVariant> mesh_v,
		 const int64_t dest,
		 const int64_t tag)
{
  if (not mesh_v) {
    throw NormalError("discrete functions are not defined on the same mesh");
  }
  std::visit(
    [&](auto&& p_mesh) {
      using MeshType = mesh_type_t<decltype(p_mesh)>;
      static const auto Dimension = MeshType::Dimension;
      using Rd           = TinyVector<Dimension>;
      if constexpr (is_polygonal_mesh_v<MeshType>) {
	std::cout<<"Serialize polygonal mesh, dim:"<<Dimension<<std::endl;


	auto costo = parallel::Messenger::getInstance().myCoupling;
	std::vector<int> shape;
	std::vector<double> pts;

	const std::shared_ptr mesh    = std::dynamic_pointer_cast<const MeshType>(p_mesh);
	const NodeValue<const Rd>& xr = mesh->xr();

	const auto& connectivity = mesh->connectivity();
	const auto node_is_owned = connectivity.nodeIsOwned().arrayView();

	Array<TinyVector<Dimension>> positions(connectivity.numberOfNodes());
	size_t sz = 0;
	for (NodeId r = 0; r < node_is_owned.size(); ++r) {
	  if (node_is_owned[r]){
	    for (unsigned short i = 0; i < Dimension; ++i) {
	      positions[sz][i] = xr[r][i];
	    }
	    sz += 1;
	  }
	}

	/*TODO: serializer directement position pour eviter une copie*/
	pts.resize(sz * Dimension);
	for (unsigned short r = 0; r < sz; ++r) {
	    for (unsigned short j = 0; j < Dimension; ++j) {
	      pts[Dimension * r + j] = positions[r][j];
	  }
	}
	shape.resize(3);
	shape[0] = sz;
	shape[1] = Dimension;
	shape[2] = 0;
	costo->sendData(shape, pts, dest, tag);

      } else {
        throw NormalError("unexpected mesh type");
      }
    },
    mesh_v->variant());
}

double Serializer::cpl_get_dt(const double& dt)
{
  std::vector<double> new_dt;
  new_dt.resize(1);
  new_dt[0] = dt*1.0;

  auto costo = parallel::Messenger::getInstance().myPyEmbeding;
  costo->Contexts->populate_context_old("cpl_get_dt", "dt", new_dt);
  costo->call_py_function("cpl_get_dt");
  return new_dt[0];
}


void
Serializer::build_graph(const std::shared_ptr<const MeshVariant> mesh_v,
			const IBoundaryDescriptor& boundary)
{

  if (not mesh_v) {
    throw NormalError("discrete functions are not defined on the same mesh");
  }
  std::visit(
    [&](auto&& p_mesh) {
      using MeshType = mesh_type_t<decltype(p_mesh)>;
      static const auto Dimension = MeshType::Dimension;
      using Rd           = TinyVector<Dimension>;
      if constexpr (is_polygonal_mesh_v<MeshType>) {
	auto costo     = parallel::Messenger::getInstance().myPyEmbeding;
	std::vector<int> shape;
	std::vector<double> pts;
	const std::shared_ptr mesh    = std::dynamic_pointer_cast<const MeshType>(p_mesh);

	const auto xr = mesh->xr();

	const auto& connectivity = mesh->connectivity();
	const auto node_is_owned = connectivity.nodeIsOwned().arrayView();

	MeshNodeBoundary mesh_node_boundary = getMeshNodeBoundary(*mesh, boundary);
	const auto node_list = mesh_node_boundary.nodeList();

	Array<TinyVector<Dimension>> positions(node_list.size());

	size_t sz = 0;
	for (size_t i_node = 0; i_node < node_list.size(); ++i_node) {
	  const NodeId r = node_list[i_node];
	  std::cout<<"Serialize node id: "<<r<<std::endl;
	  if (node_is_owned[r]){
	    for (unsigned short i = 0; i < Dimension; ++i) {
	      positions[sz][i] = xr[r][i];
	    }
	    sz += 1;
	  }
	}

	/*TODO: serializer directement position pour eviter une copie*/
	pts.resize(sz * Dimension);
	for (unsigned short r = 0; r < sz; ++r) {
	  for (unsigned short j = 0; j < Dimension; ++j) {
	    pts[Dimension * r + j] = positions[r][j];
	  }
	}
	shape.resize(3);
	shape[0] = sz;
	shape[1] = Dimension;
	shape[2] = 0;

	costo->Contexts->populate_context_old("cpl_build_graph", "shape", shape);
	costo->Contexts->populate_context_old("cpl_build_graph", "mesh", pts);
	costo->call_py_function("cpl_build_graph");
      } else {
        throw NormalError("unexpected mesh type");
      }
    },
    mesh_v->variant());
}

void Serializer::interfaces(const std::shared_ptr<const MeshVariant> mesh_v,
			    const std::vector<std::shared_ptr<const IInterfaceDescriptor>>& i_interface_list,
			    const int64_t dest,
			    const int64_t tag)
{
  if (not mesh_v) {
    throw NormalError("discrete functions are not defined on the same mesh");
  }

  std::visit(
    [&](auto&& p_mesh) {
      using MeshType = mesh_type_t<decltype(p_mesh)>;
      static const auto Dimension = MeshType::Dimension;
      using Rd           = TinyVector<Dimension>;
      if constexpr (is_polygonal_mesh_v<MeshType>) {
	std::cout<<"Serialize polygonal mesh, dim:"<<Dimension<<std::endl;


	auto costo = parallel::Messenger::getInstance().myCoupling;
	std::vector<int> shape;
	std::vector<double> pts;

	// const std::shared_ptr mesh    = std::dynamic_pointer_cast<const MeshType>(p_mesh);
	// const NodeValue<const Rd>& xr = mesh->xr();

	// const auto& connectivity = mesh->connectivity();
	// const auto node_is_owned = connectivity.nodeIsOwned().arrayView();

	// // const auto node_is_owned_generic = connectivity.isOwned<ItemType::node>();
	// // const auto node_is_owned = connectivity.nodeIsOwned();


	// Array<TinyVector<Dimension>> positions(connectivity.numberOfNodes());
	// size_t sz = 0;
	// for (NodeId r = 0; r < node_is_owned.size(); ++r) {
	//   if (node_is_owned[r]){
	//     for (unsigned short i = 0; i < Dimension; ++i) {
	//       positions[sz][i] = xr[r][i];
	//     }
	//     sz += 1;
	//   }
	// }

	// /*TODO: serializer directement position pour eviter une copie*/
	// pts.resize(sz * Dimension);
	// for (unsigned short r = 0; r < sz; ++r) {
	//     for (unsigned short j = 0; j < Dimension; ++j) {
	//       pts[Dimension * r + j] = positions[r][j];
	//   }
	// }
	// shape.resize(3);
	// shape[0] = sz;
	// shape[1] = Dimension;
	// shape[2] = 0;
	// costo->sendData(shape, pts, dest, tag);

      } else {
        throw NormalError("unexpected mesh type");
      }
    },
    mesh_v->variant());
}



// BC nodes position
void
Serializer::BC(const IBoundaryDescriptor& boundary,
	       const std::shared_ptr<const MeshVariant> mesh_v,
	       const int location,
	       const int64_t dest,
	       const int64_t tag)
{
  if (not mesh_v) {
    throw NormalError("discrete functions are not defined on the same mesh");
  }

  std::visit(
    [&](auto&& p_mesh) {
      using MeshType = mesh_type_t<decltype(p_mesh)>;
      static const auto Dimension = MeshType::Dimension;
      using Rd           = TinyVector<Dimension>;
      if constexpr (is_polygonal_mesh_v<MeshType>) {
	std::cout<<"Serialize interface of polygonal mesh, dim:"<<Dimension<<std::endl;

	auto costo = parallel::Messenger::getInstance().myCoupling;
	std::vector<int> shape;
	std::vector<double> pts;
	std::cout << "\033[01;31m"
		  << "L" << location << " ff " << at_face << "\033[00;00m" << std::endl;
	const std::shared_ptr mesh    = std::dynamic_pointer_cast<const MeshType>(p_mesh);

	if (location == at_face) {
	  std::cout << "\033[01;31m"
		    << "ICI AT FACE"
		    << "\033[00;00m" << std::endl;

	  // const MeshType& mesh_data   = *mesh->get<MeshType>();

// 	  // const auto xl = mesh->xl();
// 	  const auto xl = MeshDataManager::instance().getMeshData(mesh_data).xl();
// // ::	  // const NodeValue<const Rd>& xl = mesh_data->xl();
// // 	  // const auto xl = mesh_data.xl();

// 	  const auto& connectivity = mesh->connectivity();
// 	  const auto face_is_owned = connectivity.faceIsOwned().arrayView();

// 	  Array<TinyVector<Dimension>> fpositions(connectivity.numberOfFaces());

// 	  size_t sz = 0;
// 	  for (FaceId r = 0; r < face_is_owned.size(); ++r) {
// 	    if (face_is_owned[r]){
// 	      for (unsigned short i = 0; i < Dimension; ++i) {
// 		fpositions[sz][i] = xl[r][i];
// 	      }
// 	      sz += 1;
// 	    }
// 	  }


// 	  /*TODO: serializer directement fposition pour eviter une copie*/
// 	  MeshFaceBoundary mesh_face_boundary = getMeshFaceBoundary(*mesh, boundary);
// 	  /* mesh_face_boundary.faceList() */
// 	  const auto face_list = mesh_face_boundary.faceList();
// 	  pts.resize(face_list.size() * Dimension);

// 	  for (size_t i_face = 0; i_face < face_list.size(); ++i_face) {
// 	    const FaceId face_id = face_list[i_face];
// 	    if (face_is_owned[face_id]){

// 	      for (unsigned short j = 0; j < Dimension; ++j) {
// 		pts[Dimension * i_face + j] = fpositions[face_id][j];
// 	      }

// 	    }
// 	  }

// 	  shape.resize(3);
// 	  shape[0] = face_list.size();
// 	  shape[1] = Dimension;
// 	  shape[2] = 0;
// 	  costo->sendData(shape, pts, dest, tag);



	} else if (location == at_node) {
	  std::cout << "\033[01;31m"
		    << "ICI AT node"
		    << "\033[00;00m" << std::endl;

	  const auto xr = mesh->xr();

	  const auto& connectivity = mesh->connectivity();
	  const auto node_is_owned = connectivity.nodeIsOwned().arrayView();

	  MeshNodeBoundary mesh_node_boundary = getMeshNodeBoundary(*mesh, boundary);
	  const auto node_list = mesh_node_boundary.nodeList();

	  Array<TinyVector<Dimension>> positions(node_list.size());

	  size_t sz = 0;
	  for (size_t i_node = 0; i_node < node_list.size(); ++i_node) {
	    const NodeId r = node_list[i_node];
	    if (node_is_owned[r]){
	      for (unsigned short i = 0; i < Dimension; ++i) {
		positions[sz][i] = xr[r][i];
	      }
	      sz += 1;
	    }
	  }


	  /*TODO: serializer directement position pour eviter une copie*/
	  pts.resize(sz * Dimension);
	  for (unsigned short r = 0; r < sz; ++r) {
	    for (unsigned short j = 0; j < Dimension; ++j) {
	      pts[Dimension * r + j] = positions[r][j];
	    }
	  }
	  shape.resize(3);
	  shape[0] = sz;
	  shape[1] = Dimension;
	  shape[2] = 0;
	  costo->sendData(shape, pts, dest, tag);

	} else {
	  throw UnexpectedError(" location must be 0 or 1");
	}


      } else {
        throw NormalError("unexpected mesh type");
      }
    },
    mesh_v->variant());
}


// Field at node
void Serializer::BC_field(const std::shared_ptr<const MeshVariant> mesh_v,
			  const std::shared_ptr<const IBoundaryDescriptor>& boundary,
			  const std::shared_ptr<const ItemValueVariant>& field,
			  const int64_t dest,
			  const int64_t tag)
{
  if (not mesh_v) {
    throw NormalError("discrete functions are not defined on the same mesh");
  }

  std::visit(
    [&](auto&& p_mesh) {
      using MeshType = mesh_type_t<decltype(p_mesh)>;
      static const auto Dimension = MeshType::Dimension;
      using Rd           = TinyVector<Dimension>;
      if constexpr (is_polygonal_mesh_v<MeshType>) {
	std::cout<<"Serialize polygonal field, dim:"<<Dimension<<std::endl;

	// const std::shared_ptr i_mesh = getCommonMesh({mesh, field});
	// if (not i_mesh) {
	//   throw NormalError("primal discrete functions are not defined on the same mesh");
	// }

	auto costo = parallel::Messenger::getInstance().myCoupling;
	std::vector<int> shape;
	std::vector<double> data;

	const std::shared_ptr mesh    = std::dynamic_pointer_cast<const MeshType>(p_mesh);
	const NodeValue<const Rd>& xr = mesh->xr();

	const auto& connectivity = mesh->connectivity();
	const auto node_is_owned = connectivity.nodeIsOwned().arrayView();

	const NodeValue<const TinyVector<Dimension>> nField = field->get<NodeValue<const TinyVector<Dimension>>>();
	const auto node_list = getMeshNodeBoundary(*mesh, *boundary).nodeList();

	Array<TinyVector<Dimension>> positions(node_list.size());


	size_t sz = 0;
	for (size_t i_node = 0; i_node < node_list.size(); ++i_node) {
	  const NodeId r = node_list[i_node];
	  if (node_is_owned[r]){
	    for (unsigned short i = 0; i < Dimension; ++i) {
	      positions[sz][i] = nField[r][i];
	    }
	    sz += 1;
	  }
	}

	/*TODO: serializer directement position pour eviter une copie*/
	data.resize(sz * Dimension);
	for (unsigned short r = 0; r < sz; ++r) {
	    for (unsigned short j = 0; j < Dimension; ++j) {
	      data[Dimension * r + j] = positions[r][j];
	    }
	}

	shape.resize(3);
	shape[0] = sz;
	shape[1] = Dimension;
	shape[2] = 0;
	costo->sendData(shape, data, dest, tag);

      } else {
        throw NormalError("unexpected mesh type");
      }
    },
    mesh_v->variant());
}

#endif   // PUGS_HAS_COSTO
